# Vyřeš problém n-dam (n-queens)
#
# Jan Kybic, 2016

import time
import random

# sloupce a řádky číslujeme od 0 do 'n'
# řešení je reprezentování vektorem 'queens' sloupcových pozic,
# tady queens[i]=j znamená, že dáma je na pozici i,j

# Nejprve rekurzivní řešení postupným dosazováním
# pro n=10 potřebuje 0.01s
# pro n=20 potřebuje 23s
def safe_position(queens,i):
  """ ověř, jestli dáma v pozici (i,queens[i]) není ohrožována jinou
      v pozicích pro (j,queens[j]) pro j<i """
  for j in range(i):
    if (queens[i]==queens[j] or
        i-queens[i]==j-queens[j] or i+queens[i]==j+queens[j]):
      return False
  return True  
  
def solve_queens(n):
  """ vyřeš problém n-královen """
  # pozice 0..i-1 jsou známé, teď hledáme pozici 'i'
  def solve_internal(i):   
    nonlocal queens
    if i>=n:
      return True
    for j in range(n):
      queens[i]=j
      if safe_position(queens,i):
        if solve_internal(i+1):
          return True # řešení nalezeno
    return  False 
  queens=list(range(n))
  if solve_internal(0):
    print("Řešení nalezeno.")
    return queens
  print("Řešení nenalezeno.")

def print_queens(queens):
  n=len(queens)
  for i in range(n):
    for j in range(n):
      print('Q' if j==queens[i] else '-',end="")
    print()
      
def test_solve_queens():
#  for n in [10,20,25]:
  for n in [21,22,23,24,25]:
    t=time.clock()
    queens=solve_queens(n)
    t=time.clock()-t
    print_queens(queens)
    print("Řešení pro n=%d trvalo %5.2fs." % (n,t))

###### minimum conflict implementation

def permutation(n):
    "Create a random permutation of integers 0..n-1"
    p=list(range(n))
    for i in range(n-1):
       r=random.randrange(i,n)
       temp=p[r]
       p[r]=p[i]
       p[i]=temp
    return(p)

def conflicts(queens,i):
  """ Calculate the number of queens that (i,queens[i]) is in conflict with. """
  c=0
  for j in range(len(queens)):
    if i!=j:
      if queens[j]==queens[i] or queens[j]-j==queens[i]-i or queens[j]+j==queens[i]+i:
        c+=1
  return c

def position_costs(queens,i):
  """ Return an array 'r' such that r[j] = the number of conflicts if we reassign queen[i]=j """
  n=len(queens)
  r=[ 0 for i in range(n) ]
  for j in range(n):
    if i!=j:
      r[queens[j]]+=1 # conflicting column
      for col in [queens[j]-j+i,queens[j]+j-i]:
        if col>=0 and col<n:
          r[col]+=1
  return r
  
def solve_queens_minconflict(n,maxiter=100):
  queens=permutation(n)  # generate a random permutation
  for i in range(maxiter):
    nc=[ conflicts(queens,i) for i in range(n) ] # number of conflicts for each queen  
    ic=[ i for i in range(n) if nc[i]>0 ]  # indices of queens in conflict
    #print("iteration ",i," queens in conflict:",ic)
    #print("    queens ",queens)
    if len(ic)==0: # no conflict, puzzle is solved
      print("Solution found at iteration:",i)
      return queens
    j=ic[random.randrange(len(ic))] # choose one queen in conflict randomly
    r=position_costs(queens,j)      # costs of the positions
    #print("    costs: ",r)
    r[queens[j]]=n                  # disallow the current configuration
    minr=min(r)                     # minimum value
    ps=[ i for i in range(n) if r[i]==minr ] # indices of minimums
    p=ps[random.randrange(len(ps))] # choose one randomly
    #print("    moving queen ",j," to  ",p)
    queens[j]=p
  print("Solution not found, increase maxiter:",i)
  return None


# pro n=10  potřebuje <0.01s
# pro n=20  potřebuje 0.01s
# pro n=50  potřebuje 0.10s
# pro n=100 potřebuje 0.75s
# pro n=200 potřebuje 4.5s
# pro n=500 potřebuje 55s


def test_solve_queens_minconflict():
  for n in [10,20,50,100,200,500,1000]:
    t=time.clock()
    queens=solve_queens_minconflict(n,maxiter=1000)
    t=time.clock()-t
    if queens:
        #print_queens(queens)
        print("Řešení pro n=%d trvalo %5.2fs." % (n,t))
    else:
        print("Řešení nenalezeno, zvětšete maxiter")
          
  
