# Kód pro řešení Sudoku pomocí
# Constraint Propagation a prohledávání do hloubky
#
# Toto je mírně upravená a komentovaná verze kódu
# Petera Norviga z http://norvig.com/sudoku.html
#
# Komentáře a úpravy Jan Kybic, 2016-2017

# 9x9 políček označujeme A1 až I9
# 'unit' (jednotka) je skupina 9 políček (řada,sloupec, čtverec), která musí obsahovat
# číslice 0..9

def cross(A, B):
    "Kartézský součin dvou množin znaků."
    return [a+b for a in A for b in B]

digits   = '123456789'
rows     = 'ABCDEFGHI'
cols     = digits
squares  = cross(rows, cols)

# unitlist je seznam všech jednotek (units)

unitlist = ([cross(rows, c) for c in cols] +
            [cross(r, cols) for r in rows] +
            [cross(rs, cs) for rs in ('ABC','DEF','GHI') for cs in ('123','456','789')])

# units[políčko] je seznam všech jednotek pro dané políčko (square)
units = dict((s, [u for u in unitlist if s in u]) 
             for s in squares)

# 'peers' je mapa políček na množinu jiných políček, které jsou s ním v nějaké jednotce
peers = dict((s, set(sum(units[s],[]))-set([s]))
             for s in squares)

# stav řešení bude slovník, kde klíčem jsou jména políček a obsahem je řetězec
# číslic, které je ještě možné do daného políčka přiřadit.  Funkce 'grid_values'
# načte počáteční stav ('grid', neboli zadání sudoku) z textové reprezentace do
# reprezentace pomocí slovníku, kde klíčem jsou jména políček a obsahem číslice
# nebo '0.' (neurčené políčko). Ostatní znaky jsou
# ignorovány.

def grid_values(grid):
    "Convert grid into a dict of {square: char} with '0' or '.' for empties."
    chars = [c for c in grid if c in digits or c in '0.']
    assert len(chars) == 81
    return dict(zip(squares, chars))

# parse_grid   navíc u všech neurčených položek vloží celou přípustnou množinu číslic
# a navíc udělá propagaci podmínek - smaže z políček v rámci jednotky hodnoty, kterých určitě nemohou nabývat
  
def parse_grid(grid):
    """Convert grid to a dict of possible values, {square: digits}, or
    return False if a contradiction is detected."""
    ## To start, every square can be any digit; then assign values from the grid.
    values = dict((s, digits) for s in squares)
    for s,d in grid_values(grid).items():
        if d in digits and not assign(values, s, d):
            return False ## (Fail if we can't assign d to square s.)
    return values

# propagaci podmínek dělá funkce 'assign' - (1) když má políčko už jen jednu hodnotu, ostatní políčka v jednotce tuto hodnotu mít nemohou, (2) pokud jednotka má jen jedno možné místo pro konkrétní hodnotu, určitě tam ta hodnota bude.


def assign(values, s, d):
    """Eliminate all the other values (except d) from values[s] and propagate.
    Return values, except return False if a contradiction is detected."""
    other_values = values[s].replace(d, '')
    if all(eliminate(values, s, d2) for d2 in other_values):
        return values
    else:
        return False

def eliminate(values, s, d):
    """Eliminate d from values[s]; propagate when values or places <= 2.
    Return values, except return False if a contradiction is detected."""
    if d not in values[s]:
        return values ## Already eliminated
    values[s] = values[s].replace(d,'')
    if len(values[s]) == 0:
      return False ## Contradiction: removed last value
    ## (1) If a square s is reduced to one value d2, then eliminate d2 from the peers.
    elif len(values[s]) == 1:
        d2 = values[s]
        if not all(eliminate(values, s2, d2) for s2 in peers[s]):
            return False
    ## (2) If a unit u is reduced to only one place for a value d, then put it there.
    for u in units[s]:
      dplaces = [s for s in u if d in values[s]]
      if len(dplaces) == 0:
        return False ## Contradiction: no place for this value
      elif len(dplaces) == 1:
        # d can only be in one place in unit; assign it there
        if not assign(values, dplaces[0], d):
           return False
    return values

# funkce pro vytištění stavu řešení

def display(values):
    "Display these values as a 2-D grid."
    width = 1+max(len(values[s]) for s in squares)
    line = '+'.join(['-'*(width*3)]*3)
    for r in rows:
        print(''.join(values[r+c].center(width)+('|' if c in '36' else '')
                      for c in cols))
        if r in 'CF': print(line)
    print

# vyzkoušíme

grid1="""
..3.2.6..
9..3.5..1
..18.64..
..81.29..
7.......8
..67.82..
..26.95..
8..2.3..9
..5.1.3.."""

# toto zadání je plně vyřešeno
#display(parse_grid(grid1))

# těžší zadání nevyřešíme
grid2 = """
4.....8.5
.3.......
...7.....
.2.....6.
....8.4..
....1....
...6.3.7.
5..2.....
1.4......"""

#display(parse_grid(grid2))

# pokud existuje políčko s více možnými hodnotami, postupně je vyzkoušíme pomocí
# prohledávání do hloubky, přitom vždy aplikujeme propagaci podmínek.
# zkusím vždy nejdříve políčko s nejmenším počtem možností.


def search(values):
    "Using depth-first search and propagation, try all possible values."
    if values is False:
        return False ## Failed earlier
    if all(len(values[s]) == 1 for s in squares): 
        return values ## Solved!
    ## Chose the unfilled square s with the fewest possibilities
    n,s = min((len(values[s]), s) for s in squares if len(values[s]) > 1)
    return some(search(assign(values.copy(), s, d)) 
		             for d in values[s])

def some(seq):
    "Return some (first) element of seq that is true."
    for e in seq:
        if e: return e
    return False

def solve(grid):
  return search(parse_grid(grid))

#display(solve(grid2))
  
